import os
import sys
import json
import random
import atexit
import psutil
import hashlib
import requests
import subprocess
from packaging import version
from functools import partial
from .utils import retry_on_failure, convert_messages_wrapper, SubprocessMonitor, MaxRetriesExceededError, max_retry_wrapper  
from .cache import get_batch_cache_dir, generation_cache_wrapper
from typing import Union, List
from tqdm import tqdm

import importlib.util
flash_attn = importlib.util.find_spec("flash_attn")
if not flash_attn:
    print("Warning: flash_attn not found, recommend to install flash_attn for better performance")
    print("Simple Command: pip install flash_attn --no-build-isolation")
    print("Please refer to https://github.com/Dao-AILab/flash-attention for detailed installation instructions")

ENGINES = ["vllm", "sglang", "openai", "gemini", "mistral", "together", "claude"]
all_workers = []
verbose = False
def set_verbose(value):
    global verbose
    verbose = value
    
class WorkerInstance:
    def __init__(self, model_name, worker_addr, proc, gpu_ids=None):
        self.model_name = model_name
        self.worker_addr = worker_addr
        self.proc = proc
        self.gpu_ids = gpu_ids
    
    def __str__(self):
        return f"WorkerInstance(model_name={self.model_name}, worker_addr={self.worker_addr}, proc={self.proc}, gpu_ids={self.gpu_ids})"
    
    def __repr__(self):
        return self.__str__()


class ModelWorker:
    def __init__(
        self,
        model_name, 
        worker_addrs=None, 
        cache_dir=None, 
        use_cache=True,
        overwrite_cache=False,
        completion=False, 
        num_workers=1,
        num_gpu_per_worker=None,
        gpu_ids=None,
        dtype="auto",
        quantization=None,
        engine="vllm",
        additional_args=[],
        max_retry=None,
        verbose=False,
    ):
        """
        Return a function that calls the model worker, takes a list of messages (user, gpt, user, ...) and returns the generated text
        Args:
            model_name: model name
            worker_addrs: worker addresses, if None, launch local workers
            cache_dir: cache directory
            use_cache: use cache or not. Cache is on the hash of the input message.
            overwrite_cache: overwrite cache or not. If True, previous cache will be overwritten.
            completion: use completion or not (use chat by default)
            num_workers: number of workers
            num_gpu_per_worker: number of gpus per worker
            dtype: data type
            engine: engine name
            additional_args: additional arguments for launching the worker (vllm, sglang)
        """
        self.model_name = model_name
        self.worker_addrs = worker_addrs
        self.cache_dir = cache_dir
        self.use_cache = use_cache
        self.overwrite_cache = overwrite_cache
        self.completion = completion
        self.num_workers = num_workers
        self.num_gpu_per_worker = num_gpu_per_worker
        self.gpu_ids = gpu_ids
        self.dtype = dtype
        self.quantization = quantization
        self.engine = engine
        self.additional_args = additional_args
        self.max_retry = max_retry
        self.verbose = verbose
        self.worker_instances = [] # cuda workers instances
        self.is_sleeping = False
        
        if engine == "openai":
            from .openai_text import call_worker_openai, call_worker_openai_completion
            call_model_worker = call_worker_openai if not completion else call_worker_openai_completion
        elif engine == "gemini":
            if completion:
                raise ValueError(f"Engine {engine} does not support completion")
            from .gemini import call_worker_gemini
            call_model_worker = call_worker_gemini
        elif engine == "claude":
            if completion:
                raise ValueError(f"Engine {engine} does not support completion")
            from .claude import call_worker_claude
            call_model_worker = call_worker_claude
        elif engine == "mistral":
            if completion:
                raise ValueError(f"Engine {engine} does not support completion")
            from .mistral import call_worker_mistral
            call_model_worker = call_worker_mistral
        elif engine == "together":
            from .together import call_worker_together, call_worker_together_completion
            call_model_worker = call_worker_together if not completion else call_worker_together_completion
        elif engine == "grok":
            from .grok import call_worker_grok, call_worker_grok_completion
            call_model_worker = call_worker_grok if not completion else call_worker_grok_completion 
        elif engine == "fireworks":
            from .fireworks import call_worker_fireworks, call_worker_fireworks_completion
            call_model_worker = call_worker_fireworks if not completion else call_worker_fireworks_completion
        elif engine in ["vllm", "sglang"]:
            assert num_gpu_per_worker is not None, "num_gpu_per_worker must be provided for vllm and sglang"
            if engine == "vllm":
                from .vllm import launch_vllm_worker, call_vllm_worker, call_vllm_worker_completion
                call_worker_func = call_vllm_worker if not completion else call_vllm_worker_completion
                launch_worker_func = launch_vllm_worker
            elif engine == "sglang":
                from .sglang import launch_sglang_worker, call_sglang_worker, call_sglang_worker_completion
                call_worker_func = call_sglang_worker if not completion else call_sglang_worker_completion
                launch_worker_func = launch_sglang_worker
            else:
                raise ValueError(f"Internal error: engine {engine} not supported")
            if worker_addrs is None:
                import torch
                
                print(f"Launching model worker {model_name} locally")
                worker_addrs = []
                total_gpus = torch.cuda.device_count()
                if gpu_ids:
                    gpu_ids = [int(gpu_id) for gpu_id in gpu_ids]
                    assert len(gpu_ids) <= total_gpus, f"Error: number of gpus {len(gpu_ids)} is greater than total gpus {total_gpus}"
                    total_gpus = len(gpu_ids)
                if total_gpus < num_workers * num_gpu_per_worker:
                    if total_gpus >= num_gpu_per_worker:
                        print(f"Warning: total gpus ({total_gpus}) is less than num_workers * num_gpu_per_worker ({num_workers * num_gpu_per_worker}), using {total_gpus // num_gpu_per_worker} workers instead")
                        num_workers = total_gpus // num_gpu_per_worker
                    else:
                        print(f"Error: total gpus ({total_gpus}) is less than num_gpu_per_worker ({num_gpu_per_worker}), exiting...")
                        sys.exit(1)
                if not gpu_ids:
                    if os.environ.get("CUDA_VISIBLE_DEVICES") is not None:
                        gpus_ids = os.environ.get("CUDA_VISIBLE_DEVICES").split(",")
                        gpu_ids = [int(gpu_id) for gpu_id in gpus_ids]
                    else:
                        gpu_ids = list(range(total_gpus))
                start_port = random.randint(31000, 32000)
                for i in range(num_workers):
                    worker_addr, proc = launch_worker_func(model_name, 
                        num_gpus=num_gpu_per_worker, 
                        gpu_ids=gpu_ids[i*num_gpu_per_worker:(i+1)*num_gpu_per_worker], 
                        port=start_port+i*10,
                        dtype=dtype, quantization=quantization, additional_args=additional_args)
                    worker = WorkerInstance(model_name, worker_addr, proc, gpu_ids=gpu_ids[i*num_gpu_per_worker:(i+1)*num_gpu_per_worker])
                    worker_addrs.append(worker_addr)
                    all_workers.append(worker)
                    self.worker_instances.append(worker)
            else:
                if verbose:
                    print(f"Using existing worker at {worker_addrs}")
                if not isinstance(worker_addrs, list):
                    worker_addrs = [worker_addr]
            call_model_worker = partial(call_worker_func, worker_addrs=worker_addrs)   
            self.worker_addrs = worker_addrs
            self.gpu_ids = gpu_ids
        else:
            raise ValueError(f"Engine {engine} not supported, available engines: {ENGINES}")
        
        # wrap the call_model_worker with the model_name and other arguments
        call_model_worker = partial(call_model_worker, model_name=model_name)
        # test local worker connection
        if not completion:
            test_response = call_model_worker([{"role": "user", "content": "Hello"}], temperature=0, max_tokens=256, timeout=None)
        else:
            test_response = call_model_worker("Hello", temperature=0, max_tokens=256, timeout=None)
        if not test_response:
            print("Error: failed to connect to the worker, exiting...")
            for worker in self.worker_instances:
                cleanup_process(worker)
            sys.exit(1)
        else:
            if verbose:
                print(f"Successfully connected to the workers")
                print("Test prompt: \n", "Hello")
                print("Test response: \n", test_response)
            
        # add cache wrapper
        if use_cache:
            call_model_worker = generation_cache_wrapper(call_model_worker, model_name, cache_dir, overwrite_cache)
        else:
            if verbose:
                print("Cache is disabled")
        call_model_worker = retry_on_failure(call_model_worker, num_retries=max_retry)
        call_model_worker = convert_messages_wrapper(call_model_worker, is_completion=completion)
        set_do_cleanup(True)
        set_verbose(verbose)
        self.call_model_worker = call_model_worker
        
    def __call__(
        self,
        messages:Union[str, List[str], List[dict]],
        *args, **kwds
    ):
        if self.is_sleeping:
            print("Warning: Worker is sleeping, waking up...")
            self.wake_up_worker()
        return self.call_model_worker(messages, *args, **kwds)
    
    def sleep_worker(self, level=1):
        if self.engine in ["vllm"]:
            # vllm version should be >= 0.7.3
            from .vllm import vllm_version
            if version.parse(vllm_version) < version.parse("0.7.3"):
                raise ValueError(f"vllm version {vllm_version} does not support sleep mode, please upgrade to >= 0.7.3")
            for worker in self.worker_instances:
                response = requests.post(worker.worker_addr + "/sleep",
                                    data={"level": level})
                print(worker.worker_addr + "/sleep")
                print(response)
                assert response.status_code == 200
                print(f"Worker {worker} is sleeping")
            self.is_sleeping = True
        elif self.engine in ["sglang"]:
            raise NotImplementedError(f"Engine {self.engine} does not support sleep")
            
    def wake_up_worker(self):
        if not self.is_sleeping:
            return
        if self.engine in ["vllm"]:
            # vllm version should be >= 0.7.3
            from .vllm import vllm_version
            if version.parse(vllm_version) < version.parse("0.7.3"):
                raise ValueError(f"vllm version {vllm_version} does not support sleep mode, please upgrade to >= 0.7.3")
            for worker in self.worker_instances:
                response = requests.post(worker.worker_addr + "/wake_up")
                assert response.status_code == 200
                print(f"Worker {worker} is woken up")
        elif self.engine in ["sglang"]:
            raise NotImplementedError(f"Engine {self.engine} does not support wake up")
    
    def __str__(self):
        return f"ModelWorker(model_name={self.model_name}, num_workers={self.num_workers}, num_gpu_per_worker={self.num_gpu_per_worker}, worker_addrs={self.worker_addrs}, cache_dir={self.cache_dir}, use_cache={self.use_cache}, overwrite_cache={self.overwrite_cache}, completion={self.completion}, dtype={self.dtype}, quantization={self.quantization}, engine={self.engine}, additional_args={self.additional_args}, max_retry={self.max_retry}, verbose={self.verbose})"
    
    def __repr__(self):
        return self.__str__()


do_cleanup = True
def set_do_cleanup(value):
    global do_cleanup
    do_cleanup = value
    
def kill_process_and_children(pid):
    # check if the process is still alive
    if 'psutil' not in sys.modules:
        # possibly the main process is in the final stage of termination, no need to kill the child processes
        return None
    try:
        parent = psutil.Process(pid)
    except psutil.NoSuchProcess:
        print(f"No process with PID {pid} found.")
        return

    children = parent.children(recursive=True)
    for child in children:
        try:
            child.kill()
            # print(f"Killed child process {child.pid}")
        except psutil.NoSuchProcess:
            print(f"Child process {child.pid} already terminated.")
    
    try:
        parent.kill()
        # print(f"Killed parent process {pid}")
    except psutil.NoSuchProcess:
        print(f"Parent process {pid} already terminated.")
    return True
        
def cleanup_process(worker:Union[ModelWorker, WorkerInstance, SubprocessMonitor, subprocess.Popen]):
    if isinstance(worker, ModelWorker):
        for worker_instance in worker.worker_instances:
            cleanup_process(worker_instance)
        return
    if isinstance(worker, WorkerInstance):
        proc = worker.proc.proc
    elif isinstance(worker, SubprocessMonitor):
        proc = worker.proc
    elif isinstance(worker, subprocess.Popen):
        proc = worker
    else:
        raise ValueError(f"Unknown process type {type(proc)}")
    killed = kill_process_and_children(proc.pid)
    if verbose and killed:
        print(f"Model Worker terminated: {worker} ")
    return killed

@atexit.register
def cleanup_all_workers():
    if not do_cleanup:
        return
    for worker in all_workers:
        cleanup_process(worker)
    if all_workers and verbose:
        print("All workers terminated.")
    all_workers.clear()


class LLMEngine:
    
    def __init__(self, verbose=False, num_gpus: int = None, gpu_ids: Union[List[int], str] = None):
        self.workers = []
        self.loaded_model_worker = {}
        import torch
        self.verbose = verbose
        total_gpus = torch.cuda.device_count()
        if gpu_ids:
            assert isinstance(gpu_ids, (list, str)), "passed gpu_ids must be a list or a string"
            if isinstance(gpu_ids, str):
                gpu_ids = [int(gpu_id) for gpu_id in gpu_ids.split(",")]
            assert all(isinstance(gpu_id, int) for gpu_id in gpu_ids), "passed gpu_ids must be a list of integers"
            assert len(gpu_ids) <= total_gpus, f"Error: passed gpu_ids {gpu_ids} is greater than total gpus {total_gpus}"
        else:
            if os.environ.get("CUDA_VISIBLE_DEVICES") is not None:
                gpu_ids = [int(gpu_id) for gpu_id in os.environ.get("CUDA_VISIBLE_DEVICES").split(",")]
            else:
                gpu_ids = list(range(total_gpus))
            if num_gpus is not None:
                num_gpus = int(num_gpus)
                assert num_gpus <= total_gpus, f"Error: passed num_gpus {num_gpus} is greater than total gpus {total_gpus}"
                gpu_ids = gpu_ids[:num_gpus]
        self.gpu_ids = gpu_ids
        self.num_gpus = len(gpu_ids)
        if verbose:
            print(f"LLMEngine initialized with {self.num_gpus} GPUs: {gpu_ids}")
        
    def get_available_gpu_ids(self):
        worker_used_gpu_ids = []
        for worker in self.workers:
            for worker_instance in worker.worker_instances:
                worker_used_gpu_ids.extend(worker_instance.gpu_ids)
        available_gpu_ids = [gpu_id for gpu_id in self.gpu_ids if gpu_id not in worker_used_gpu_ids]
        return available_gpu_ids
    
    def load_model(
        self,
        model_name,
        worker_addrs=None,
        cache_dir=None,
        use_cache=True,
        overwrite_cache=False,
        completion=False,
        num_workers=1,
        num_gpu_per_worker=None,
        dtype="auto",
        quantization=None,
        engine="vllm",
        additional_args=[],
        max_retry=None,
        verbose=None
    ):
        """
        Load a model
        Args:
            model_name: model name
            worker_addrs: worker addresses, if None, launch local workers
            cache_dir: cache directory
            use_cache: use cache or not. Cache is on the hash of the input message.
            overwrite_cache: overwrite cache or not. If True, previous cache will be overwritten.
            completion: use completion or not (use chat by default)
            num_workers: number of workers
            num_gpu_per_worker: number of gpus per worker
            dtype: data type
            engine: engine name
            additional_args: additional arguments for launching the worker (vllm, sglang)
            max_retry: maximum number of retries
            verbose: verbose
        """
        verbose = self.verbose or verbose
        if self.workers:
            print("Warning: previous workers are not cleaned up, please call unload_model() to clean up previous workers")
        self.model_name = model_name
        available_gpu_ids = self.get_available_gpu_ids()
        if engine in ["vllm", "sglang"]:
            if not num_gpu_per_worker:
                print("Warning: num_gpu_per_worker not provided, using 1 GPU per worker")
                num_gpu_per_worker = 1
            num_required_gpus = num_workers * num_gpu_per_worker
            if len(available_gpu_ids) < num_required_gpus:
                print("Error: No available GPU to launch the model worker")
                print("Provided GPU IDs for this LLMEngine class: ", self.gpu_ids)
                print("Used GPU IDs for all workers: ", [worker_instance.gpu_ids for worker in self.workers for worker_instance in worker.worker_instances])
                print("Available GPU IDs: ", available_gpu_ids)
                print("Number of required GPUs: ", num_required_gpus)
                raise ValueError("Not enought available GPU to launch the model worker")
            gpu_ids = available_gpu_ids[:num_required_gpus]
        else:
            num_required_gpus = num_workers
            gpu_ids = None
            
        model_worker = ModelWorker(
            model_name, 
            worker_addrs=worker_addrs, 
            cache_dir=cache_dir, 
            use_cache=use_cache,
            overwrite_cache=overwrite_cache,
            completion=completion, 
            num_workers=num_workers,
            num_gpu_per_worker=num_gpu_per_worker,
            gpu_ids=gpu_ids,
            dtype=dtype,
            quantization=quantization,
            engine=engine,
            additional_args=additional_args,
            max_retry=max_retry,
            verbose=verbose,
        )
        self.workers.append(model_worker)
        self.loaded_model_worker[model_name] = model_worker
        return model_worker

    def sleep_model(
        self,
        model_name,
        level=1
    ):
        model_worker = self.loaded_model_worker.get(model_name)
        model_worker.sleep_worker(level=level)
        
    def wake_up_model(
        self,
        model_name,
    ):
        model_worker = self.loaded_model_worker.get(model_name)
        model_worker.wake_up_worker()
    
    def call_model(
        self, 
        model_name, 
        messages:Union[List[str], List[dict], str],
        timeout:int=60, 
        conv_system_msg=None, 
        **generate_kwargs
    ):
        """
        Call a model
        Args:
            model_name: model name
            messages: list of messages in openai format
            timeout: timeout
            conv_system_msg: conversation system message
            generate_kwargs: generation arguments
        """
        call_model_worker = self.loaded_model_worker.get(model_name)
        if call_model_worker is None:
            raise ValueError(f"Model {model_name} not loaded, please call load_model() first")
        try:
            return call_model_worker(messages, timeout=timeout, conv_system_msg=conv_system_msg, **generate_kwargs)
        except MaxRetriesExceededError as e:
            print(e)
            return None
    
    def batch_call_model(
        self,
        model_name,
        batch_messages:List[Union[List[str], List[dict], str]],
        timeout:int=60,
        conv_system_msg=None,
        num_proc=8,
        desc=None,
        disable_batch_api=False,
        max_batch_size=None,
        **generate_kwargs
    ):
        """
        Batch call a model
        Args:
            model_name: model name
            batch_messages: list of list of messages in openai format or list of strings
            timeout: timeout
            conv_system_msg: conversation system message
            num_proc: number of processes
            generate_kwargs: generation arguments
        """
        supported_batch_api_engines = ["openai", "claude"]
        model_worker = self.loaded_model_worker.get(model_name)
        engine = model_worker.engine
        if engine not in supported_batch_api_engines or disable_batch_api:
            if model_worker is None:
                raise ValueError(f"Model {model_name} not loaded, please call load_model() first")
            
            batch_cache_dir = get_batch_cache_dir(model_name, None)
            to_write_batch_messages = [
                {"input": message, "generation_kwargs": generate_kwargs} for message in batch_messages
            ]
            hash_str = hashlib.md5("".join([
                str(x) for x in to_write_batch_messages
            ]).encode()).hexdigest()
            hash_result_file = batch_cache_dir / f"{hash_str}_batch_results.jsonl"
            if hash_result_file.exists():
                results = []
                with open(hash_result_file, "r") as f:
                    for line in f:
                        message = json.loads(line)
                        results.append(message["output"])
            else:
                from functools import partial
                from multiprocessing import Pool
                num_proc = min(num_proc, len(batch_messages))
                if model_worker.is_sleeping:
                    print("Warning: Worker is sleeping, waking up...")
                    model_worker.wake_up_worker()
                call_model_worker_mp = partial(model_worker.call_model_worker, timeout=timeout, conv_system_msg=conv_system_msg, **generate_kwargs)
                call_model_worker_mp = partial(max_retry_wrapper, call_model_worker_mp)
                with Pool(num_proc) as p:
                    results = list(tqdm(p.imap(call_model_worker_mp, batch_messages), total=len(batch_messages), desc=desc or "LLMEngine Batch Inference"))
                if results:
                    for i, message in enumerate(to_write_batch_messages):
                        message["output"] = results[i]
                    with open(batch_cache_dir / f"{hash_str}_batch_results.jsonl", "w") as f:
                        for message in to_write_batch_messages:
                            f.write(json.dumps(message) + "\n")
        else:
            if engine == "openai":
                print("Using OpenAI batch API")
                from .openai_text import openai_batch_request, check_batch_status, get_batch_progress, get_batch_result
                batch_request_func = openai_batch_request
            elif engine == "claude":
                print("Using Claude batch API")
                from .claude import claude_batch_request, check_batch_status, get_batch_progress, get_batch_result
                batch_request_func = claude_batch_request
            else:
                raise ValueError(f"Engine {engine} not supported for batch API")
            if max_batch_size is None:
                results = batch_request_func(model_name, batch_messages, conv_system_msg=conv_system_msg, desc=desc, **generate_kwargs)
            else:
                # using multiprocess to submit batch request per batch
                from functools import partial
                import time
                
                max_slots = num_proc
                all_batch_inputs = [
                    batch_messages[i:i+max_batch_size] for i in range(0, len(batch_messages), max_batch_size)
                ]
                # submit detach jobs
                batch_ids = [None] * len(all_batch_inputs)
                # wait for all jobs to finish and periodically check the status
                idx = 0
                tqdm_bar = tqdm(total=len(batch_ids), desc=desc or "LLMEngine Batch Inference")
                all_batch_status = [check_batch_status(batch_id)['status'] if batch_id is not None else "pending" for batch_id in batch_ids]
                while True:
                    batch_id = batch_ids[idx]
                    if batch_id is None:
                        cur_slots = len([bstatus for bstatus in all_batch_status if bstatus not in ['pending', 'completed', 'cancelled', 'canceled', "failed", "expired"]])
                        if cur_slots < max_slots:
                            batch_id = batch_request_func(model_name, all_batch_inputs[idx], conv_system_msg=conv_system_msg, desc=desc, **generate_kwargs, detach=True)
                            batch_ids[idx] = batch_id
                        else:
                            tqdm_bar.n = 0
                            tqdm_bar.total = len(all_batch_inputs[idx])
                            tqdm_bar.desc = "pending" + f" (batch {idx+1}/{len(batch_ids)})"
                            tqmd_postfix = tqdm_postfix = {
                                "completed": 0,
                                "total": len(all_batch_inputs[idx]),
                                "failed": 0,
                            }
                            tqdm_bar.set_postfix_str(tqmd_postfix)
                            tqdm_bar.refresh()
                            idx = (idx + 1) % len(batch_ids)
                            time.sleep(5)
                            continue
                        
                    n, total, tqdm_postfix, cur_batch_status = get_batch_progress(batch_id)
                    tqdm_bar.set_postfix_str(tqdm_postfix)
                    tqdm_bar.n = n
                    tqdm_bar.total = total
                    tqdm_bar.desc = cur_batch_status + f" (batch {idx+1}/{len(batch_ids)})"
                    tqdm_bar.refresh()
                    all_batch_status[idx] = cur_batch_status
                    if all(status == "completed" for status in all_batch_status):
                        tqdm_bar.close()
                        break
                    idx = (idx + 1) % len(batch_ids)
                    time.sleep(10)
                
                # collect results
                all_batch_results = []
                for i, batch_id in enumerate(batch_ids):
                    batch_results = get_batch_result(batch_id, generate_kwargs=generate_kwargs)
                    if batch_results:
                        all_batch_results.extend(batch_results)
                    else:
                        raise ValueError(f"Warning: batch {batch_id} has no results")
                results = all_batch_results
                    
        return results
    
    def __call__(self, *args, **kwds):
        return self.call_model(*args, **kwds)
                    
    def unload_model(self, model_name=None):
        to_remove_local_workers = []
        to_remove_global_workers = []
        for worker in self.workers:
            if model_name is None or worker.model_name == model_name:
                print(f"Unloading model worker: {worker}")
                cleanup_process(worker) 
                if worker in all_workers:
                    to_remove_global_workers.append(worker)
                if worker in self.workers:
                    to_remove_local_workers.append(worker)
        for worker in to_remove_global_workers:
            all_workers.remove(worker)
        for worker in to_remove_local_workers:
            self.workers.remove(worker)
        
    def __del__(self):
        pass
        
        


from llm_engines import LLMEngine
from PIL import Image
import requests
from io import BytesIO
response = requests.get("https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg")
image = Image.open(BytesIO(response.content)).resize((256, 256))
image.save("./test.jpg")
messages_with_image = [
    {
        "role": "user",
        "content": [
            {
                "type": "text",
                "text": "What's in the image?"
            },
            {
                "type": "image",
                "image": image
            }
        ]
    }
] # the 'image' type is not offical format of openai API, LLM-Engines will convert it into image_url type internally
messages_with_image_url = [
    {
        "role": "user",
        "content": [
            {
                "type": "text",
                "text": "What's in the image?"
            },
            {
                "type": "image_url",
                "image_url": {"url": "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg"}
            }
        ]
    }
] # the 'image_url' type is the offical format of openai API
additional_args=[]
# engine="openai"; model_name="gpt-4o-mini"
# engine="claude"; model_name="claude-3-5-sonnet-20241022"
# engine="gemini"; model_name="gemini-2.0-flash"
# engine="grok"; model_name="grok-2-vision-latest"
# engine="sglang"; model_name="meta-llama/Llama-3.2-11B-Vision-Instruct"; additional_args=["--chat-template=llama_3_vision"] # refer to 
engine="vllm"; model_name="microsoft/Phi-3.5-vision-instruct"; additional_args=["--limit-mm-per-prompt", "image=2", "--max-model-len", "4096"] # refer to vllm serve api
llm = LLMEngine()
llm.load_model(
    model_name=model_name, 
    engine=engine, # or "vllm", "together", "mist
    use_cache=False,
    additional_args=additional_args,
)
response = llm.call_model(model_name, messages_with_image, temperature=0.0, max_tokens=None)
print(response)
response = llm.call_model(model_name, messages_with_image_url, temperature=0.0, max_tokens=None)
print(response)


import os
import hashlib
import base64
from copy import deepcopy
from typing import List, Union
from cachetools import LRUCache
from functools import lru_cache
from pathlib import Path

@lru_cache(maxsize=None)
def get_cache_file(model_name_or_path, cache_dir):
    model_name = model_name_or_path.split("/")[-2:]
    model_name = "/".join(model_name)
    if cache_dir is not None:
        cache_file = Path(cache_dir) / f"{model_name}.jsonl"
    else:
        cache_file = Path(os.path.expanduser(f"~/llm_engines/generation_cache/{model_name}.jsonl"))
    if not cache_file.parent.exists():
        cache_file.parent.mkdir(parents=True)
    return cache_file

@lru_cache(maxsize=None)
def get_batch_cache_dir(model_name_or_path, cache_dir):
    model_name = model_name_or_path.split("/")[-2:]
    model_name = "/".join(model_name)
    if cache_dir is not None:
        batch_cache_dir = Path(cache_dir) / f"{model_name}_batch_cache"
    else:
        batch_cache_dir = Path(os.path.expanduser(f"~/llm_engines/generation_cache/{model_name}_batch_cache"))
    if not batch_cache_dir.exists():
        batch_cache_dir.mkdir(parents=True)
    return batch_cache_dir

MAX_PRINTABLE_IMAGE_URL_LENGTH = 100
def get_printable_messages(messages):
    # mainly for image_url, only keep the first 100 characters
    messages = deepcopy(messages)
    for message in messages:
        if isinstance(message["content"], list):
            for sub_message in message["content"]:
                if sub_message["type"] == "image_url":
                    sub_message["image_url"]["url"] = sub_message["image_url"]["url"][:MAX_PRINTABLE_IMAGE_URL_LENGTH] + \
                        (f"... ({len(sub_message['image_url']['url']) - MAX_PRINTABLE_IMAGE_URL_LENGTH} more characters)" if len(sub_message['image_url']['url']) > MAX_PRINTABLE_IMAGE_URL_LENGTH else "")
                elif sub_message["type"] == "image":
                    sub_message["image"] = sub_message["image"][:100]
    return messages

def get_inputs_hash(inputs:Union[str, List[dict]], conv_system_msg, generate_kwargs=None):
    
    inputs = inputs.copy()
    if isinstance(inputs, str):
        try:
            return hashlib.md5(inputs.encode()).hexdigest()
        except UnicodeEncodeError as e:
            return hashlib.md5(inputs.encode('utf-16', 'surrogatepass').decode('utf-16').encode('utf-8')).hexdigest()
    
    # inputs is a list of dicts in openai format
    
    if conv_system_msg:
        to_hash_messages = [{
            "role": "system",
            "content": conv_system_msg
        }] + inputs
    else:
        to_hash_messages = inputs
    
    to_hash_inputs = []
    for message in to_hash_messages:
        role = message["role"]
        content = message["content"]
        if isinstance(content, str):
            to_hash_inputs.append(f"{role}:{content}")
        elif isinstance(content, list):
            strs = []
            for sub_content in content:
                if sub_content["type"] == "text":
                    strs.append(sub_content["text"])
                elif sub_content["type"] == "image_url":
                    if "url" not in sub_content["image_url"]:
                        raise ValueError("image_url must have a url key")
                    image_url_hash = hashlib.md5(sub_content["image_url"]["url"].encode()).hexdigest()
                    strs.append(f"image_url:{image_url_hash}")
                else:
                    raise ValueError(f"Unknown content type {sub_content['type']}")
            to_hash_inputs.append(f"{role}:{''.join(strs)}")
        else:
            raise ValueError(f"Unknown content type {type(content)}")
    
    if generate_kwargs:
        to_hash_inputs.append(str(generate_kwargs))
    
    try:
        return hashlib.md5("".join(to_hash_inputs).encode()).hexdigest()
    except UnicodeEncodeError as e:
        return hashlib.md5("".join(to_hash_inputs).encode('utf-16', 'surrogatepass').decode('utf-16').encode('utf-8')).hexdigest()
    
    
    



import subprocess
import time
import os
import signal
import os
import signal
import json
import traceback
import threading
import openai
import inspect
import base64
import datetime
import regex as re
import requests
from io import BytesIO
from PIL import Image
from pathlib import Path
from typing import Union, List
from typing import List
from transformers import AutoTokenizer
from functools import partial
from .cache import load_cache, get_inputs_hash, get_cache_file

default_gen_params = {
    "temperature": 0.0,
    "max_tokens": None,
    "top_p": 1.0,
    "timeout": 600,
}
class SubprocessMonitor:
    def _monitor(self):
        while True:
            if self.proc.poll() is not None:
                print("Subprocess has exited with code", self.proc.returncode)
                os.kill(os.getpid(), signal.SIGTERM)  # Exit the main process
                break
            time.sleep(5)
            
    def __init__(self, command, **kwargs):
        print("Launching subprocess with command:\n", " ".join(command))
        self.proc = subprocess.Popen(command, **kwargs)
        # self.monitor_thread = threading.Thread(target=self._monitor)
        # self.monitor_thread.start()
    
class ChatTokenizer:
    def __init__(self, model_name):
        
        self.model_name = model_name
        self.tokenizer = AutoTokenizer.from_pretrained(model_name, trust_remote_code=True)
        self.system_message = None
        try:
            self.max_length = self.tokenizer.model_max_length
        except AttributeError:
            self.max_length = 4096
        if not isinstance(self.max_length, int):
            self.max_length = 4096
        if self.max_length > 1e6:
            self.max_length = 1e6
            
        if self.tokenizer.chat_template:
            self.apply_chat_template = self.apply_chat_template_default
            print("Using hugging face chat template for model", model_name)
            self.chat_template_source = "huggingface"
        else:
            self.apply_chat_template = None
            self.chat_template_source = None
        print("Example prompt: \n", self.example_prompt())
        
    def apply_chat_template_default(
        self, 
        messages:List[str],
        add_generation_prompt:bool=True,
        chat_template:str=None
    ):
        prompt = self.tokenizer.apply_chat_template(
            messages, 
            add_generation_prompt=add_generation_prompt,
            tokenize=False,
            chat_template=chat_template,
        )
        return prompt
    
    def example_prompt(self):
        if not self.apply_chat_template:
            return "Chat template not available for this model"
        else:
            example_messages = [
                {"role": "user", "content": "Hello"},
                {"role": "assistant", "content": "Hi"},
                {"role": "user", "content": "How are you?"},
                {"role": "assistant", "content": "I'm good, how about you?"},
            ]
            return self.apply_chat_template(example_messages)
    
    def __call__(self, messages:List[str], **kwargs):
        if not self.apply_chat_template:
            raise NotImplementedError("Chat template not available for this model")
        return self.apply_chat_template(messages, **kwargs)


def convert_messages(messages:Union[List[str], List[dict], str]):
    """
    Convert messages to the format expected by the model
    """
    if all(isinstance(item, dict) for item in messages):
        assert all("content" in item for item in messages), "content key not found in messages"
        assert all("role" in item for item in messages), "role key not found in messages"
        if messages[0]["role"] == "system":
            conv_system_msg = messages[0]["content"]
            messages = messages[1:]
        else:
            conv_system_msg = None
        new_messages = messages
    else:
        if isinstance(messages, str):
            messages = [messages]
        assert all(isinstance(item, str) for item in messages)
        new_messages = []
        for i, message in enumerate(messages):
            if i % 2 == 0:
                new_messages.append({"role": "user", "content": message})
            else:
                new_messages.append({"role": "assistant", "content": message})
        conv_system_msg = None
    
    # assert the correct format of images
    for message in new_messages:
        assert "role" in message, "role key not found in message"
        assert "content" in message, "content key not found in message"
        if isinstance(message["content"], str):
            pass
        elif isinstance(message["content"], list):
            for sub_content in message["content"]:
                assert sub_content["type"] in sub_content, f"'{sub_content['type']}' key not found in sub_content of type {sub_content['type']}"
                if sub_content["type"] == "text":
                    assert isinstance(sub_content["text"], str), "text key not found in sub_content"
                elif sub_content["type"] == "image_url":
                    assert "url" in sub_content["image_url"] and isinstance(sub_content["image_url"]["url"], str), "url key not found in sub_content['image_url']"
                elif sub_content["type"] == "image":
                    assert isinstance(sub_content["image"], Image.Image), "The image key in of 'image' type must be a PIL Image object"
                    # change image to image_url
                    sub_content["type"] = "image_url"
                    sub_content["image_url"] = {"url": encode_base64_image_url(sub_content["image"])}
                    del sub_content["image"]
                else:
                    raise ValueError(f"Unsupported sub_content type: {sub_content['type']}")
        else:
            raise ValueError(f"Unsupported content type: {type(message['content'])}")
    
    return new_messages, conv_system_msg

def _convert_messages_wrapper(messages:Union[List[str], List[dict], str], call_model_worker, is_completion=False, **generate_kwargs):
    if not is_completion:
        messages, conv_system_msg = convert_messages(messages)
        generate_kwargs["conv_system_msg"] = conv_system_msg
    else:
        assert isinstance(messages, str), "Completion model only accepts a single string input"
    # add default generation parameters
    for key, value in default_gen_params.items():
        if key not in generate_kwargs:
            generate_kwargs[key] = value
    return call_model_worker(messages, **generate_kwargs)

def convert_messages_wrapper(call_model_worker, is_completion=False):
    return partial(_convert_messages_wrapper, call_model_worker=call_model_worker, is_completion=is_completion)

class MaxRetriesExceededError(Exception):
    pass

short_error_instances = [
    openai.BadRequestError,
]

def _retry_on_failure(*args, call_model_worker=None, num_retries=5, **kwargs):
    if not call_model_worker:
        raise ValueError("call_model_worker is required")
    try:
        return call_model_worker(*args, **kwargs)
    except Exception as e:
        if not num_retries:
            if any(isinstance(e, error_instance) for error_instance in short_error_instances):
                print(str(e))
            else:
                print(traceback.format_exc())
            raise MaxRetriesExceededError(f"Max retries exceeded for call_model_worker (num_retries={num_retries})")
        for i in range(num_retries):
            try:
                return call_model_worker(*args, **kwargs)
            except Exception as e:
                print("Error in call_model_worker, retrying... (Error: {})".format(e))
                time.sleep(1)
                if i >= num_retries - 1 and not isinstance(e, TimeoutError):
                    if any(isinstance(e, error_instance) for error_instance in short_error_instances):
                        print(str(e))
                    else:
                        # format dump of the last error and
                        print(traceback.format_exc())
        raise MaxRetriesExceededError(f"Max retries exceeded for call_model_worker (num_retries={num_retries})")
    
def retry_on_failure(call_model_worker, num_retries=5):
    return partial(_retry_on_failure, call_model_worker=call_model_worker, num_retries=num_retries)

def timeout_handler(signum, frame):
    raise TimeoutError("Function call timed out")

def with_timeout(timeout):
    def decorator(func):
        def wrapper(*args, **kwargs):
            result = [TimeoutError(f"Function call timed out (timeout={timeout})")]
            stop_event = threading.Event()

            def target():
                try:
                    result[0] = func(*args, **kwargs)
                except Exception as e:
                    result[0] = e

            thread = threading.Thread(target=target)
            thread.start()
            thread.join(timeout)
            if thread.is_alive():
                stop_event.set()
                raise TimeoutError(f"Function call timed out (timeout={timeout})")
            if isinstance(result[0], Exception):
                raise result[0]
            return result[0]
        return wrapper
    return decorator

def max_retry_wrapper(func, *args, **kwargs):
    try:
        return func(*args, **kwargs)
    except MaxRetriesExceededError as e:
        print(str(e))
        return None

def get_function_arg_names(func):
    signature = inspect.signature(func)
    parameters = signature.parameters
    
    arg_names = []
    kwarg_names = []
    
    for name, param in parameters.items():
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            arg_names.append(f"*{name}")
        elif param.kind == inspect.Parameter.VAR_KEYWORD:
            kwarg_names.append(f"**{name}")
        else:
            arg_names.append(name)
    
    return arg_names, kwarg_names

def encode_base64_image(image:Image.Image, image_format="PNG") -> str:
    im_file = BytesIO()
    image.save(im_file, format=image_format)
    im_bytes = im_file.getvalue()
    im_64 = base64.b64encode(im_bytes).decode("utf-8")
    return im_64

def encode_base64_image_url(image:Image.Image, image_format="PNG") -> str:
    return f"data:image/{image_format};base64,{encode_base64_image(image, image_format)}"

def decode_base64_image_url(base64_uri:str) -> Image.Image:
    # Split the URI to get the base64 data
    try:
        # Remove the "data:image/format;base64," prefix
        header, base64_data = base64_uri.split(',', 1)
        # Get image format from header
        image_format = header.split('/')[1].split(';')[0]
        # Decode base64 string
        image_data = base64.b64decode(base64_data)
        # Create image from binary data
        image = Image.open(BytesIO(image_data))
        return image
    except Exception as e:
        raise ValueError(f"Failed to decode base64 image: {str(e)}")
    
def is_base64_image_url(base64_uri:str) -> bool:
    return base64_uri.startswith("data:image/")

def load_image(image_path:str) -> Image.Image:
    # either http or local file path
    if image_path.startswith("http"):
        response = requests.get(image_path)
        image = Image.open(BytesIO(response.content)).convert("RGB")
    else:
        image = Image.open(image_path).convert("RGB")
    return image


import os
import json
import hashlib
import time
import filelock
import random
import openai
from datetime import datetime
from openai import OpenAI
from typing import List, Union
from pathlib import Path
from tqdm import tqdm


batch_submission_status_file = Path(os.path.expanduser(f"~/llm_engines/generation_cache")) / "openai_batch_cache" / "batch_submission_status.json"
batch_submission_status_file.parent.mkdir(parents=True, exist_ok=True)

def read_batch_submission_status():
    print("Loading batch submission status from", batch_submission_status_file)
    if batch_submission_status_file.exists():
        
        lock = filelock.FileLock(str(batch_submission_status_file) + ".lock", timeout=30)
        try:
            with lock:
                with open(batch_submission_status_file, "r") as f:
                    batch_submission_status = json.load(f)
        except filelock.Timeout as e:
            print("Timeout acquiring lock")
            raise e
    else:
        batch_submission_status = {}
    return batch_submission_status

def write_batch_submission_status(batch_submission_status):
    lock = filelock.FileLock(str(batch_submission_status_file) + ".lock", timeout=30)
    try:
        with lock:
            with open(batch_submission_status_file, "w") as f:
                json.dump(batch_submission_status, f, indent=4)
    except filelock.Timeout as e:
        print("Timeout acquiring lock")
        raise e
            
        
# no image, multi-turn, do not use openai_generate, but can refer to it
def call_worker_openai(messages:List[str], model_name, timeout:int=60, conv_system_msg=None, **generate_kwargs) -> str:
    # change messages to openai format
    if conv_system_msg:
        new_messages = [{"role": "system", "content": conv_system_msg}] + messages
    else:
        new_messages = messages
    # initialize openai client
    client = OpenAI()
    o_series = ["o1", "o3"]
    if any(o in model_name for o in o_series):
        # fixed parameters for openai o1 models
        max_tokens = generate_kwargs.pop("max_tokens", None)
        if max_tokens is not None:
            generate_kwargs["max_completion_tokens"] = max_tokens
        generate_kwargs['temperature'] = 1
        generate_kwargs['top_p'] = 1
        generate_kwargs['n'] = 1
        generate_kwargs['frequency_penalty'] = 0
        generate_kwargs['presence_penalty'] = 0
    # call openai
    completion = client.chat.completions.create(
        model=model_name,
        messages=new_messages,
        timeout=timeout,
        **generate_kwargs,
    )
    stream = generate_kwargs.get("stream", False)
    
    if "logprobs" in generate_kwargs:
        return_logprobs = True
    
    if not stream:
        if "logprobs" not in generate_kwargs or not generate_kwargs["logprobs"]:
            if len(completion.choices) > 1:
                return [c.message.content for c in completion.choices]
            else:
                return completion.choices[0].message.content
        else:
            if len(completion.choices) > 1:
                return [c.message.content for c in completion.choices], [c.logprobs.dict() for c in completion.choices]
            else:
                return completion.choices[0].message.content, completion.choices[0].logprobs.dict()
    else:
        def generate_stream():
            for chunk in completion:
                if chunk.choices[0].delta.content is not None:
                    yield chunk.choices[0].delta.content
        return generate_stream()

def call_worker_openai_completion(prompt:str, model_name, timeout:int=60, **generate_kwargs) -> str:
    # initialize openai client
    client = OpenAI()
    # call openai
    completion = client.completions.create(
        model=model_name,
        prompt=prompt,
        timeout=timeout,
        **generate_kwargs,
    )
    stream = generate_kwargs.get("stream", False)
    if not stream:
        if len(completion.choices) > 1:
            return [c.text for c in completion.choices]
        else:
            return completion.choices[0].text
    else:
        def generate_stream():
            for chunk in completion:
                if chunk.choices[0].text is not None:
                    yield chunk.choices[0].text
        return generate_stream()
    
def save_batch_file(batch_messages:List[Union[str, dict]], model:str, batch_name:str=None, cache_dir=None, custom_ids=None, **generate_kwargs):
    if isinstance(batch_messages[0], str):
        batch_messages = [{"role": "user", "content": message} for message in batch_messages]

    if cache_dir is None:
        cache_dir = Path(os.path.expanduser(f"~/llm_engines/generation_cache")) / "openai_batch_cache"
    else:
        cache_dir = Path(cache_dir) / "openai_batch_cache"
    cache_dir.mkdir(parents=True, exist_ok=True)
    if custom_ids is None:
        custom_ids = [f"request-{i+1}" for i in range(len(batch_messages))]
    assert len(custom_ids) == len(batch_messages)
    if batch_name is None:
        # batch_name = f"{model}_{datetime.now().strftime('%Y%m%d_%H%M%S')}"
        hash_str = "".join([str(message) for message in batch_messages])
        batch_name = f"{model}_{hashlib.md5(hash_str.encode()).hexdigest()}"
    file_path = cache_dir / f"{batch_name}.jsonl"
    with open(file_path, "w") as f:
        for custom_id, message in zip(custom_ids, batch_messages):
            if isinstance(message, str):
                message = {"role": "user", "content": message}
            f.write(json.dumps({"custom_id": custom_id, "method": "POST", "url": "/v1/chat/completions", "body": {"model": model, "messages": message, **generate_kwargs}}, ensure_ascii=False) + "\n")
    return file_path

def submit_batch_file(batch_file:str, output_path:str=None, project_name:str=None, description:str=None):
    """
    Submit a batch of queries to OpenAI API
    Args:
        batch_file: str, path to the batch jsonl file
        output_path: str, path to the output file; if not provided, will be saved in the same directory as the batch_file, ending with ".batch_results.jsonl"
        project_name: str, project name, default to "vapo", keyword to filter batches
        description: str, description of the batch, default to the batch_file name, can also be used to filter batches
    Returns:
        batch_result_id: str, the id of the batch submission
    """
    if isinstance(batch_file, str):
        batch_file = Path(batch_file)
        
    # internally maintain a batch submission status json
    batch_submission_status = read_batch_submission_status()
    
    client = OpenAI()
    batch_input_file = client.files.create(
        file=open(batch_file, "rb"),
        purpose="batch"
    )
    if not project_name:
        project_name = "llm_engines"
    description = description if description is not None else batch_file.stem
    output_path = output_path if output_path is not None else batch_file.parent / f"{description}.batch_results.jsonl"
    batch_input_file_id = batch_input_file.id
    
    batch_file_hash = hashlib.md5(batch_file.read_bytes()).hexdigest()
    batch_file_size = batch_file.stat().st_size
    
    for key, value in batch_submission_status.items():
        value_input_file_metadata = value["input_path_metadata"]
        if not os.path.samefile(value["input_path"], batch_file):
            # print(f"Batch {key} has a different input file. need to resubmit.")
            continue
        if batch_file_size != value_input_file_metadata["size"]:
            # print(f"Batch {key} has a newer version of the input file. need to resubmit.")
            continue
        if batch_file_hash != value_input_file_metadata["hash"]:
            # print(f"Batch {key} has a different input hash. need to resubmit.")
            continue
        if value['status'] in ["validating", "in_progress", "finalizing", "completed"]:
            print(f"Batch {key} is still in progress. Skipping submission.")
            return key
        else:
            continue
    
    batch_result = None
    for batch in client.batches.list(limit=10):
        if batch.metadata and (batch.metadata.get('input_path') == str(batch_file)) and batch.status in ["validating", "in_progress", "finalizing", "completed"]:
            batch_result = batch
            break
    
    completion_window = "24h"
    endpoint = "/v1/chat/completions"
    if batch_result is None:
        batch_result = client.batches.create(
            input_file_id=batch_input_file_id,
            endpoint=endpoint,
            completion_window=completion_window,
            metadata={
                "project": project_name,
                "description": description,
                "input_path": str(batch_file),
                "output_path": str(output_path)
            }
        )
        print(f"Batch {batch_result.id} submitted")
        submit_time = str(datetime.now())
    else:
        print(f"Batch already exists for {batch_file}, but not found in the managing file, writing to {batch_submission_status_file}")
        submit_time = batch_result.created_at
    
    # time should be in the current timezone, in the format like 2022-01-01T00:00:00
    batch_submission_status[batch_result.id] = {    
        "project": project_name,
        "description": description,
        "endpoint": endpoint,
        "completion_window": completion_window,
        "input_path": str(batch_file),
        "input_path_metadata": {
            "hash": hashlib.md5(batch_file.read_bytes()).hexdigest(),
            "size": batch_file.stat().st_size
        },
        "output_path": str(output_path),
        "batch_input_file_id": batch_input_file_id,
        "batch_result_id": batch_result.id,
        "status": batch_result.status,
        "timeline": {
            "submitted": submit_time,
            "completed": None,
            "failed": None,
            "downloaded": None
        },
        "last_updated": str(datetime.now()),
        "openai_batch_metadata": batch_result.to_dict()
    }
    
    write_batch_submission_status(batch_submission_status)
    return batch_result.id

def check_batch_status(batch_id, overwrite:bool=False):
    # internally maintain a batch submission status json
    batch_submission_status = read_batch_submission_status()
    if batch_id in batch_submission_status:
        batch_status = batch_submission_status[batch_id]["status"]
    else:
        client = OpenAI()
        try:
            batch = client.batches.retrieve(batch_id)
        except openai.error.NotFoundError:
            print(f"Batch {batch_id} not found.")
            return None
    if batch_status == "completed":
        output_path = Path(batch_submission_status[batch_id]["output_path"])
        if output_path.exists() and not overwrite:
            print(f"Output file {output_path} already exists. Skipping writing.")
        else:
            print(f"Downloading output file for batch {batch_id}")
            client = OpenAI()
            batch = client.batches.retrieve(batch_id)
            batch_id = batch.id
            batch_status = batch.status
            if batch.metadata is not None and "output_path" in batch.metadata:
                batch_output_path = batch.metadata["output_path"]
            else:
                batch_output_path = f"./batch_results/{batch_id}.batch_results.jsonl"
                batch_submission_status[batch_id]["output_path"] = batch_output_path
            content = client.files.content(batch.output_file_id)
            output_path = batch_output_path
            output_path = Path(output_path)
            output_path.parent.mkdir(parents=True, exist_ok=True)
            if output_path.exists() and overwrite:
                print(f"Overwriting file {output_path}")
            content.write_to_file(output_path)
            print(f"Output file written to {output_path}")
            batch_submission_status[batch_id]["status"] = "completed"
            batch_submission_status[batch_id]["timeline"]["completed"] = str(datetime.now())
            batch_submission_status[batch_id]["timeline"]["downloaded"] = str(datetime.now())
            batch_submission_status[batch_id]["openai_batch_metadata"].update(batch.to_dict())
    else:
        client = OpenAI()
        batch = client.batches.retrieve(batch_id)
        batch_id = batch.id
        batch_status = batch.status
        batch_desc = batch.metadata["description"] if batch.metadata is not None and "description" in batch.metadata else ""
        batch_project_name = batch.metadata["project"] if batch.metadata is not None and "project" in batch.metadata else ""
        if batch.metadata is not None and "output_path" in batch.metadata:
            batch_output_path = batch.metadata["output_path"]
        else:
            batch_output_path = f"./batch_results/{batch_id}.batch_results.jsonl"
            batch_submission_status[batch_id]["output_path"] = batch_output_path
        # print(f"{batch_id: <20} {batch_status: <20} {batch_project_name: <20} {batch_desc: <20}")
        if batch_status == "failed":
            print(f"Batch {batch_id} failed.")
            batch_submission_status[batch_id]["status"] = "failed"
            batch_submission_status[batch_id]["timeline"]["failed"] = str(datetime.now())
            batch_submission_status[batch_id]["openai_batch_metadata"].update(batch.to_dict())
        else:
            batch_submission_status[batch_id]["status"] = batch_status
            batch_submission_status[batch_id]["openai_batch_metadata"].update(batch.to_dict())
            
    batch_submission_status[batch_id]["last_updated"] = str(datetime.now())
    write_batch_submission_status(batch_submission_status)
    return batch_submission_status[batch_id]

def get_batch_progress(batch_id):
    batch_status = check_batch_status(batch_id)
    num_completed = batch_status["openai_batch_metadata"]['request_counts']['completed']
    num_total = batch_status["openai_batch_metadata"]['request_counts']['total']
    num_failed = batch_status["openai_batch_metadata"]['request_counts']['failed']
    n = num_completed
    total = num_total
    tqdm_postfix = {
        "completed": num_completed,
        "total": num_total,
        "failed": num_failed
    }
    return n, total, tqdm_postfix, batch_status['status']

def get_batch_result(batch_id, generate_kwargs={}):
    batch_status = check_batch_status(batch_id)
    if not batch_status["status"] == "completed":
        return None
    output_path = batch_status["output_path"]
        
    if not os.path.exists(output_path):
        client = OpenAI()
        batch = client.batches.retrieve(batch_id)
        content = client.files.content(batch.output_file_id)
        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        print(f"Downloading output file for batch {batch_id}")
        content.write_to_file(output_path)
        
    results = []
    with open(output_path, "r") as f:
        results = [json.loads(line) for line in f.readlines()]
    if "logprobs" not in generate_kwargs or not generate_kwargs["logprobs"]:
        all_completions = [[choice['message']['content'] for choice in x['response']['body']['choices']] for x in results]
    else:
        all_completions = [[(choice['message']['content'], choice['logprobs']) for choice in x['response']['body']['choices']] for x in results]
    if all(len(x) == 1 for x in all_completions):
        all_completions = [x[0] for x in all_completions]
    results = all_completions
    return results

def openai_batch_request(
    model_name:str,
    batch_messages:List[Union[str, List[str], List[dict]]],
    conv_system_msg:str=None,
    desc:str=None,
    detach:bool=False,
    **generate_kwargs
):
    if isinstance(batch_messages[0], str):
        batch_messages = [[{"role": "user", "content": message}] for message in batch_messages]
    elif isinstance(batch_messages[0], list):
        if isinstance(batch_messages[0][0], str):
            batch_messages = [[{
                "role": "user" if i % 2 == 0 else "assistant",
                "content": message
            } for i, message in enumerate(messages)] for messages in batch_messages]
        elif isinstance(batch_messages[0][0], dict):
            assert all("role" in message for message in batch_messages[0]), "Error: role key not found in the message"
            assert all("content" in message for message in batch_messages[0]), "Error: content key not found in the message"
        else:
            raise ValueError("Error: unknown message format")
    else:
        raise ValueError("Error: unknown message format")
    if conv_system_msg:
        batch_messages = [[{"role": "system", "content": conv_system_msg}] + messages for messages in batch_messages]
    if "stream" in generate_kwargs:
        generate_kwargs.pop("stream")
    batch_file = save_batch_file(batch_messages, model_name, **generate_kwargs)
    batch_result_id = submit_batch_file(batch_file)
    if detach:
        return batch_result_id
    num_total = len(batch_messages)
    tqdm_bar = tqdm(total=num_total, desc=desc or "LLMEngine Batch Inference")
    while True:
        batch_status = check_batch_status(batch_result_id)
        assert batch_status is not None, f"Error: {batch_result_id} not found in batch submission status or OpenAI API"
        num_completed = batch_status["openai_batch_metadata"]['request_counts']['completed']
        num_total = batch_status["openai_batch_metadata"]['request_counts']['total']
        num_failed = batch_status["openai_batch_metadata"]['request_counts']['failed']
        tqdm_bar.n = num_completed
        tqdm_bar.total = num_total
        tqdm_bar.set_postfix(completed=num_completed, total=num_total, failed=num_failed)
        if batch_status["status"] == "completed":
            tqdm_bar.close()
            break
        elif batch_status["status"] == "finalizing":
            tqdm_bar.desc = "Finalizing"
            tqdm_bar.refresh()
        elif batch_status["status"] == "validating":
            tqdm_bar.desc = "Validating"
            tqdm_bar.refresh()
        elif batch_status["status"] == "failed":
            tqdm_bar.close()
            print("Batch failed")
            break
        elif batch_status["status"] == "in_progress":
            tqdm_bar.desc = "In Progress"
            tqdm_bar.refresh()
        elif batch_status["status"] == "cancelled":
            tqdm_bar.close()
            print("Batch cancelled")
            break
        elif batch_status["status"] == "expired":
            tqdm_bar.close()
            print("Batch expired")
            break
        elif batch_status["status"] == "cancelling":
            tqdm_bar.desc = "Cancelling"
            tqdm_bar.refresh()
        else:
            tqdm_bar.desc = batch_status["status"]
            tqdm_bar.refresh()
        time.sleep(random.randint(5, 10))
    
    results = get_batch_result(batch_result_id, generate_kwargs)
    return results

if __name__ == "__main__":
    from icecream import ic
    ic(call_worker_openai(["Hello"], "gpt-3.5-turbo"))
    ic(call_worker_openai_completion("Hello", "gpt-3.5-turbo-instruct"))


import json
import os
import time
from functools import partial
from typing import Union, List
from pathlib import Path
from .cache_sqlite3 import load_cache as load_sqlite3_cache
from .cache_lru import load_cache as load_lru_cache
from .cache_dict import load_cache as load_dict_cache
from .cache_utils import get_cache_file, get_inputs_hash, get_printable_messages, get_batch_cache_dir

load_cache = load_sqlite3_cache

def _generation_cache_wrapper(inputs: Union[str, List[dict]], call_model_worker, model_name, cache_dir=None, overwrite_cache=False, **generate_kwargs):
    cache_dict = load_cache(model_name, cache_dir)
    
    conv_system_msg = generate_kwargs.get("conv_system_msg", "")
    if "n" in generate_kwargs:
        non_hash_keys = ["timeout", "stream"]
        inputs_hash = get_inputs_hash(inputs, conv_system_msg, {k: v for k, v in generate_kwargs.items() if k not in non_hash_keys})
    else:
        inputs_hash = get_inputs_hash(inputs, conv_system_msg)
    
    if not overwrite_cache:
        cached_value = cache_dict[inputs_hash]
        if cached_value:
            if "logprobs" not in generate_kwargs or not generate_kwargs["logprobs"]:
                return cached_value["output"]
            elif "logprobs" in cached_value:
                return cached_value["output"], cached_value["logprobs"]
    
    response = call_model_worker(inputs, **generate_kwargs)
    if isinstance(response, tuple):
        generated_text, logprobs = response
    else:
        generated_text = response
        logprobs = None
    cache_item = {
        "input": get_printable_messages(inputs),
        "output": generated_text,
        "logprobs": logprobs,
        "model_name": model_name,
        'tstamp': time.time(),
        "time": time.strftime('%Y-%m-%d %H:%M:%S', time.localtime()),
        "generate_kwargs": generate_kwargs
    }
    
    # cache_dict[inputs_hash] = cache_item
    
    cache_file = get_cache_file(model_name, cache_dir)
    with open(cache_file, "a+") as f:
        f.write(json.dumps({inputs_hash: cache_item}) + "\n")
    
    return response

def generation_cache_wrapper(call_model_worker, model_name, cache_dir=None, overwrite_cache=False):
    print(f"Using efficient multi-level cache for model {model_name}")
    if cache_dir is None:
        env_cache_dir = os.getenv("LLM_ENGINES_CACHE_DIR")
        if env_cache_dir:
            cache_dir = Path(env_cache_dir)
        else:
            cache_dir = Path(os.path.expanduser(f"~/llm_engines/generation_cache"))
    print(f"Cache directory: {cache_dir}")
    load_cache(model_name, cache_dir) # preload cache
    
    return partial(_generation_cache_wrapper, call_model_worker=call_model_worker, model_name=model_name, cache_dir=cache_dir, overwrite_cache=overwrite_cache)


import os
from typing import List, Union
from fireworks.client import Fireworks

# no image, multi-turn
def call_worker_fireworks(messages:List[str], model_name, timeout:int=60, conv_system_msg=None, **generate_kwargs) -> str:
    # change messages to openai format
    if conv_system_msg:
        new_messages = [{"role": "system", "content": conv_system_msg}] + messages
    else:
        new_messages = messages
    # initialize openai client
    client = Fireworks(api_key=os.environ["FIREWORKS_API_KEY"])
    # call fireworks
    completion = client.chat.completions.create(
        model=model_name,
        messages=new_messages,
        **generate_kwargs,
    )
    stream = generate_kwargs.get("stream", False)
    
    if "logprobs" in generate_kwargs:
        return_logprobs = True
    
    if not stream:
        if "logprobs" not in generate_kwargs or not generate_kwargs["logprobs"]:
            if len(completion.choices) > 1:
                return [c.message.content for c in completion.choices]
            else:
                return completion.choices[0].message.content
        else:
            if len(completion.choices) > 1:
                return [c.message.content for c in completion.choices], [c.logprobs.dict() for c in completion.choices]
            else:
                return completion.choices[0].message.content, completion.choices[0].logprobs.dict()
    else:
        def generate_stream():
            for chunk in completion:
                if chunk.choices[0].delta.content is not None:
                    yield chunk.choices[0].delta.content
        return generate_stream()

def call_worker_fireworks_completion(prompt:str, model_name, timeout:int=60, **generate_kwargs) -> str:
    # initialize openai client
    client = Fireworks(api_key=os.environ["FIREWORKS_API_KEY"])
    # call fireworks
    print(generate_kwargs)
    if "max_tokens" not in generate_kwargs:
        generate_kwargs["max_tokens"] = 256 # have to set max_tokens to be explicit
    completion = client.completions.create(
        model=model_name,
        prompt=prompt,
        **generate_kwargs,
    )
    stream = generate_kwargs.get("stream", False)
    if not stream:
        if len(completion.choices) > 1:
            return [c.text for c in completion.choices]
        else:
            return completion.choices[0].text
    else:
        def generate_stream():
            for chunk in completion:
                if chunk.choices[0].text is not None:
                    yield chunk.choices[0].text
        return generate_stream()

if __name__ == "__main__":
    from icecream import ic
    generate_kwargs = {
        "top_p": 1,
        "top_k": 40,
        "presence_penalty": 0,
        "frequency_penalty": 0,
        "temperature": 0.6,
        "max_tokens": 20480
    }
    ic(call_worker_fireworks(["Hello"], "accounts/fireworks/models/deepseek-r1", **generate_kwargs))
    ic(call_worker_fireworks_completion("Hello", "accounts/fireworks/models/deepseek-r1", **generate_kwargs))


import os
import json
import hashlib
import time
import filelock
import random
import openai
from datetime import datetime
from openai import OpenAI
from typing import List, Union
from pathlib import Path
from tqdm import tqdm
from .cache import get_printable_messages

# no image, multi-turn, do not use grok_generate, but can refer to it
def call_worker_grok(messages:List[str], model_name, timeout:int=60, conv_system_msg=None, **generate_kwargs) -> str:
    # change messages to openai format
    if conv_system_msg:
        new_messages = [{"role": "system", "content": conv_system_msg}] + messages
    else:
        new_messages = messages
    # initialize openai client
    client = OpenAI(api_key=os.environ["XAI_API_KEY"], base_url="https://api.x.ai/v1")
    # call grok
    completion = client.chat.completions.create(
        model=model_name,
        messages=new_messages,
        timeout=timeout,
        **generate_kwargs,
    )
    stream = generate_kwargs.get("stream", False)
    
    if "logprobs" in generate_kwargs:
        return_logprobs = True
    
    if not stream:
        if "logprobs" not in generate_kwargs or not generate_kwargs["logprobs"]:
            if len(completion.choices) > 1:
                return [c.message.content for c in completion.choices]
            else:
                return completion.choices[0].message.content
        else:
            if len(completion.choices) > 1:
                return [c.message.content for c in completion.choices], [c.logprobs.dict() for c in completion.choices]
            else:
                return completion.choices[0].message.content, completion.choices[0].logprobs.dict()
    else:
        def generate_stream():
            for chunk in completion:
                if chunk.choices[0].delta.content is not None:
                    yield chunk.choices[0].delta.content
        return generate_stream()

def call_worker_grok_completion(prompt:str, model_name, timeout:int=60, **generate_kwargs) -> str:
    # initialize openai client
    client = OpenAI(api_key=os.environ["XAI_API_KEY"], base_url="https://api.x.ai/v1")
    # call grok
    if "max_tokens" not in generate_kwargs:
        generate_kwargs["max_tokens"] = 256 # have to set max_tokens to be explicit
    completion = client.completions.create(
        model=model_name,
        prompt=prompt,
        timeout=timeout,
        **generate_kwargs,
    )
    stream = generate_kwargs.get("stream", False)
    if not stream:
        if len(completion.choices) > 1:
            return [c.text for c in completion.choices]
        else:
            return completion.choices[0].text
    else:
        def generate_stream():
            for chunk in completion:
                if chunk.choices[0].text is not None:
                    yield chunk.choices[0].text
        return generate_stream()
    
if __name__ == "__main__":
    from icecream import ic
    ic(call_worker_grok(["Hello"], "grok-2-latest"))
    ic(call_worker_grok_completion("Hello", "grok-2-latest"))


import anthropic
import os
import requests
import json
import hashlib
import time
import filelock
import random
import regex as re
from copy import deepcopy
from typing import List, Union
from anthropic import NOT_GIVEN
from datetime import datetime
from pathlib import Path
from tqdm import tqdm
from .utils import with_timeout, is_base64_image_url, encode_base64_image, load_image, decode_base64_image_url
batch_submission_status_file = Path(os.path.expanduser(f"~/llm_engines/generation_cache")) / "claude_batch_cache" / "batch_submission_status.json"
batch_submission_status_file.parent.mkdir(parents=True, exist_ok=True)


def read_batch_submission_status():
    print("Loading batch submission status from", batch_submission_status_file)
    if batch_submission_status_file.exists():
        
        lock = filelock.FileLock(str(batch_submission_status_file) + ".lock", timeout=30)
        try:
            with lock:
                with open(batch_submission_status_file, "r") as f:
                    batch_submission_status = json.load(f)
        except filelock.Timeout as e:
            print("Timeout acquiring lock")
            raise e
    else:
        batch_submission_status = {}
    return batch_submission_status

def write_batch_submission_status(batch_submission_status):
    batch_submission_status_file.parent.mkdir(parents=True, exist_ok=True)
    lock = filelock.FileLock(str(batch_submission_status_file) + ".lock", timeout=30)
    try:
        with lock:
            with open(batch_submission_status_file, "w") as f:
                json.dump(batch_submission_status, f, indent=4)
    except filelock.Timeout as e:
        print("Timeout acquiring lock")
        raise e

# 5MB
MAX_IMAGE_SIZE = 5 * 1024 * 1024
def preprocess_claude_messages(messages:List[dict]) -> List[dict]:
    messages = deepcopy(messages)
    for message in messages:
        if isinstance(message['content'], list):
            for sub_message in message['content']:
                if sub_message['type'] == "image_url":
                    if is_base64_image_url(sub_message['image_url']['url']):
                        # if image size is greater than 5MB, decode and resize and re-encode
                        im64 = sub_message['image_url']['url'].split(",", 1)[1]
                        current_size = len(im64)
                        # print("current_size", current_size)
                        if current_size > MAX_IMAGE_SIZE:
                            print("Warning: Image size is greater than 5MB. Resizing image due to Claude API limit.")
                            image = decode_base64_image_url(sub_message['image_url']['url'])
                            image_format = image.format if image.format else "png"
                            scale_factor = (MAX_IMAGE_SIZE / current_size) ** 0.6
                            image = image.resize((int(image.width * scale_factor), int(image.height * scale_factor)))
                            im64 = encode_base64_image(image, image_format)
                            media_type = "image/png"
                        else:
                            start_idx = sub_message['image_url']['url'].find("image/")
                            end_idx = sub_message['image_url']['url'].find(";base64")
                            media_type = sub_message['image_url']['url'][start_idx:end_idx].lower()
                    else:
                        image = load_image(sub_message['image_url']['url'])
                        current_size = image.size[0] * image.size[1] * 3
                        if current_size > MAX_IMAGE_SIZE:
                            print("Warning: Image size is greater than 5MB. Resizing image due to Claude API limit.")
                            scale_factor = (MAX_IMAGE_SIZE / current_size) ** 0.6
                            image = image.resize((int(image.size[0] * scale_factor), int(image.size[1] * scale_factor)))
                        image_format = image.format if image.format else "png"
                        im64= encode_base64_image(image, image_format)
                        media_type = f"image/{image_format}".lower()
                    sub_message['source'] = {
                        "type": "base64",
                        "media_type": media_type,
                        "data": im64
                    }
                    sub_message['type'] = "image"
                    sub_message.pop('image_url')
    return messages
            
# no image, multi-turn, do not use openai_generate, but can refer to it
def call_worker_claude(messages:List[str], model_name, timeout:int=60, conv_system_msg=None, **generate_kwargs) -> str:
    # change messages to mistral format
    client = anthropic.Anthropic(api_key=os.environ.get("ANTHROPIC_API_KEY"))
    # change messages to openai format
    if conv_system_msg:
        new_messages = [{"role": "system", "content": conv_system_msg}] + messages
    else:
        new_messages = messages
    new_messages = preprocess_claude_messages(new_messages)
             
    generate_kwargs.pop("n", None) # claude does not have n
    if not generate_kwargs.get("max_tokens", None):
        generate_kwargs["max_tokens"] = 1024
    stream = generate_kwargs.pop("stream", False)
    if "logprobs" in generate_kwargs:
        raise ValueError("Error: logprobs is not supported in claude")
    @with_timeout(timeout)
    def get_response():
        completion = client.messages.create(
            model=model_name,
            messages=new_messages,
            system=conv_system_msg if conv_system_msg else NOT_GIVEN,
            timeout=timeout,
            **generate_kwargs,
        )
        if len(completion.content) > 1:
            return [c.text for c in completion.content]
        else:
            return completion.content[0].text
        
    @with_timeout(timeout)
    def stream_response():
        with client.messages.stream(
            model=model_name,
            messages=new_messages,
            system=conv_system_msg if conv_system_msg else NOT_GIVEN,
            timeout=timeout,
            **generate_kwargs,
        ) as stream:
            for text in stream.text_stream:
                yield text
    
    if not stream:
        return get_response()
    else:
        return stream_response()

def save_batch_file(batch_messages:List[Union[str, dict]], model:str, batch_name:str=None, cache_dir=None, custom_ids=None, **generate_kwargs):
    if isinstance(batch_messages[0], str):
        batch_messages = [{"role": "user", "content": message} for message in batch_messages]

    if cache_dir is None:
        cache_dir = Path(os.path.expanduser(f"~/llm_engines/generation_cache")) / "openai_batch_cache"
    else:
        cache_dir = Path(cache_dir) / "openai_batch_cache"
    cache_dir.mkdir(parents=True, exist_ok=True)
    if custom_ids is None:
        custom_ids = [f"request-{i+1}" for i in range(len(batch_messages))]
    assert len(custom_ids) == len(batch_messages)
    if batch_name is None:
        # batch_name = f"{model}_{datetime.now().strftime('%Y%m%d_%H%M%S')}"
        hash_str = "".join([str(message) for message in batch_messages])
        batch_name = f"{model}_{hashlib.md5(hash_str.encode()).hexdigest()}"
    file_path = cache_dir / f"{batch_name}.jsonl"
    with open(file_path, "w") as f:
        for custom_id, message in zip(custom_ids, batch_messages):
            if isinstance(message, str):
                message = {"role": "user", "content": message}
            f.write(json.dumps({"custom_id": custom_id, "params": {"model": model, "messages": message, **generate_kwargs}}, ensure_ascii=False) + "\n")
    return file_path

def submit_batch_file(batch_file:str, output_path:str=None, project_name:str=None, description:str=None):
    """
    Submit a batch of queries to OpenAI API
    Args:
        batch_file: str, path to the batch jsonl file
        output_path: str, path to the output file; if not provided, will be saved in the same directory as the batch_file, ending with ".batch_results.jsonl"
        project_name: str, project name, default to "vapo", keyword to filter batches
        description: str, description of the batch, default to the batch_file name, can also be used to filter batches
    Returns:
        batch_result_id: str, the id of the batch submission
    """
    if isinstance(batch_file, str):
        batch_file = Path(batch_file)
        
    # internally maintain a batch submission status json
    batch_submission_status = read_batch_submission_status()
        
    
    client = anthropic.Anthropic()
    with open(batch_file, "r") as f:
        batch_inputs = [json.loads(line) for line in f.readlines()]
    if not project_name:
        project_name = "llm_engines"
    description = description if description is not None else batch_file.stem
    output_path = output_path if output_path is not None else batch_file.parent / f"{description}.batch_results.jsonl"
    
    batch_file_hash = hashlib.md5(batch_file.read_bytes()).hexdigest()
    batch_file_size = batch_file.stat().st_size
    
    for key, value in batch_submission_status.items():
        value_input_file_metadata = value["input_path_metadata"]
        if not os.path.samefile(value["input_path"], batch_file):
            # print(f"Batch {key} has a different input file. need to resubmit.")
            continue
        if batch_file_size != value_input_file_metadata["size"]:
            # print(f"Batch {key} has a newer version of the input file. need to resubmit.")
            continue
        if batch_file_hash != value_input_file_metadata["hash"]:
            # print(f"Batch {key} has a different input hash. need to resubmit.")
            continue
        if value['status'] in ["errored", "expired", "canceled"]:
            continue
        print(f"Batch {key} is still in progress. Skipping submission.")
        return key
    
    
    batch_result = client.beta.messages.batches.create(
        requests=batch_inputs,
    )
    print(f"Batch {batch_result.id} submitted")
    
    claude_batch_metadata = batch_result.to_dict()
    for key in claude_batch_metadata:
        if isinstance(claude_batch_metadata[key], datetime):
            claude_batch_metadata[key] = str(claude_batch_metadata[key])
    # time should be in the current timezone, in the format like 2022-01-01T00:00:00
    batch_submission_status[batch_result.id] = {    
        "project": project_name,
        "description": description,
        "endpoint": None,
        "completion_window": None,
        "input_path": str(batch_file),
        "input_path_metadata": {
            "hash": hashlib.md5(batch_file.read_bytes()).hexdigest(),
            "size": batch_file.stat().st_size,
            "total": len(batch_inputs)
        },
        "output_path": str(output_path),
        "batch_input_file_id": None,
        "batch_result_id": batch_result.id,
        "status": "submitted",
        "timeline": {
            "submitted": str(datetime.now()),
            "completed": None,
            "failed": None,
            "downloaded": None,
        },
        "last_updated": str(datetime.now()),
        "claude_batch_metadata": claude_batch_metadata
    }
    
    write_batch_submission_status(batch_submission_status)
    return batch_result.id

def check_batch_status(batch_result_id, overwrite:bool=False):
    batch_submission_status = read_batch_submission_status()
    client = anthropic.Anthropic(api_key=os.environ.get("ANTHROPIC_API_KEY"))
    try:
        batch = client.beta.messages.batches.retrieve(batch_result_id)
    except anthropic.NotFoundError as e:
        return None
    batch_id = batch.id
    batch_status = batch.processing_status
    batch_desc = batch_submission_status[batch_id]["description"] if batch_id in batch_submission_status else ""
    batch_project_name = batch_submission_status[batch_id]["project"] if batch_id in batch_submission_status else ""
    if batch_submission_status[batch_id]["output_path"]:
        batch_output_path = batch_submission_status[batch_id]["output_path"]
    else:
        batch_output_path = f"./batch_results/{batch_id}.batch_results.jsonl"
        batch_submission_status[batch_id]["output_path"] = batch_output_path
    # print(f"{batch_id: <20} {batch_status: <20} {batch_project_name: <20} {batch_desc: <20}")
    claude_batch_metadata = batch.to_dict()
    for key in claude_batch_metadata:
        if isinstance(claude_batch_metadata[key], datetime):
            claude_batch_metadata[key] = str(claude_batch_metadata[key])
    if batch_status == "ended":
        output_path = batch_output_path
        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        if output_path.exists() and not overwrite:
            print(f"File {output_path} already exists. Skipping writing to file.")
        else:
            if output_path.exists() and overwrite:
                print(f"Overwriting file {output_path}")
                
            # retrieve the results
            print(f"Downloading output file for batch {batch_id}")
            results = client.beta.messages.batches.results(batch_id)
            # save url to file via web requests, download the results_url file, and save to output_path
            with open(output_path, "wb") as f:
                for result in results:
                    f.write(json.dumps(result.to_dict(), ensure_ascii=False).encode() + b"\n")
            print(f"Output file written to {output_path}")
        with open(output_path, "r") as f:
            results = [json.loads(line) for line in f.readlines()]
        all_result_types = [x['result']['type'] for x in results]
        if all(all_result_types) == "errored" or all(all_result_types) == "error":
            batch_status = "errored"
            batch_submission_status[batch_id]["status"] = "errored"
            batch_submission_status[batch_id]["timeline"]["failed"] = str(datetime.now())
        elif all(all_result_types) == "canceled":
            batch_status = "canceled"
            batch_submission_status[batch_id]["status"] = "canceled"
            batch_submission_status[batch_id]["timeline"]["canceled"] = str(datetime.now())
        elif all(all_result_types) == "expired":
            batch_status = "expired"
            batch_submission_status[batch_id]["status"] = "expired"
            batch_submission_status[batch_id]["timeline"]["expired"] = str(datetime.now())
        else:
            batch_submission_status[batch_id]["status"] = "completed"
        batch_submission_status[batch_id]["timeline"]["completed"] = str(datetime.now())
        batch_submission_status[batch_id]["timeline"]["downloaded"] = str(datetime.now())
        batch_submission_status[batch_id]["last_updated"] = str(datetime.now())
        batch_submission_status[batch_id]["claude_batch_metadata"].update(claude_batch_metadata)
    else:
        batch_submission_status[batch_id]["status"] = batch_status
        batch_submission_status[batch_id]["last_updated"] = str(datetime.now())
        batch_submission_status[batch_id]["claude_batch_metadata"].update(claude_batch_metadata)
    write_batch_submission_status(batch_submission_status)
    return batch_submission_status[batch_id]

def get_batch_progress(batch_result_id):
    batch_status = check_batch_status(batch_result_id)
    num_succeeded = batch_status["claude_batch_metadata"]['request_counts']['succeeded']
    num_processing = batch_status["claude_batch_metadata"]['request_counts']['processing']
    num_errored = batch_status["claude_batch_metadata"]['request_counts']['errored']
    num_expired = batch_status["claude_batch_metadata"]['request_counts']['expired']
    num_canceled = batch_status["claude_batch_metadata"]['request_counts']['canceled']
    num_total = num_succeeded + num_processing + num_errored + num_expired + num_canceled
    n = num_succeeded + num_errored + num_expired + num_canceled
    total = num_total
    tqdm_postfix = {
        "completed": num_succeeded,
        "processing": num_processing,
        "errored": num_errored,
        "expired": num_expired,
        "canceled": num_canceled
    }
    return n, total, tqdm_postfix, batch_status["status"]

def get_batch_result(batch_id, generate_kwargs={}):
    batch_status = check_batch_status(batch_id)
    if batch_status["status"] == "completed":
        output_path = batch_status["output_path"]
        results = []
        with open(output_path, "r") as f:
            raw_results = [json.loads(line) for line in f.readlines()]
        results = []
        for item in raw_results:
            if item["result"]["type"] == "succeeded":
                results.append(item["result"]['message']["content"][0]['text'])
            else:
                results.append(None)
        print("Batch requests status counts:", batch_status["claude_batch_metadata"]['request_counts'])
    else:
        results = None
    return results

def claude_batch_request(
    model_name:str,
    batch_messages:List[Union[str, List[str], List[dict]]],
    conv_system_msg:str=None,
    desc:str=None,
    detach:bool=False,
    **generate_kwargs
):
    if isinstance(batch_messages[0], str):
        batch_messages = [[{"role": "user", "content": message}] for message in batch_messages]
    elif isinstance(batch_messages[0], list):
        if isinstance(batch_messages[0][0], str):
            batch_messages = [[{
                "role": "user" if i % 2 == 0 else "assistant",
                "content": message
            } for i, message in enumerate(messages)] for messages in batch_messages]
        elif isinstance(batch_messages[0][0], dict):
            assert all("role" in message for message in batch_messages[0]), "Error: role key not found in the message"
            assert all("content" in message for message in batch_messages[0]), "Error: content key not found in the message"
        else:
            raise ValueError("Error: unknown message format")
    else:
        raise ValueError("Error: unknown message format")
    if conv_system_msg:
        batch_messages = [[{"role": "system", "content": conv_system_msg}] + messages for messages in batch_messages]
    if "stream" in generate_kwargs:
        generate_kwargs.pop("stream")
    batch_file = save_batch_file(batch_messages, model_name, **generate_kwargs)
    batch_result_id = submit_batch_file(batch_file)
    if detach:
        return batch_result_id
    num_total = len(batch_messages)
    tqdm_bar = tqdm(total=num_total, desc=desc or "LLMEngine Batch Inference")
    while True:
        batch_status = check_batch_status(batch_result_id)
        assert batch_status is not None, f"Error: batch {batch_result_id} not found in Anthropic API"
        num_succeeded = batch_status["claude_batch_metadata"]['request_counts']['succeeded']
        num_processing = batch_status["claude_batch_metadata"]['request_counts']['processing']
        num_errored = batch_status["claude_batch_metadata"]['request_counts']['errored']
        num_expired = batch_status["claude_batch_metadata"]['request_counts']['expired']
        num_canceled = batch_status["claude_batch_metadata"]['request_counts']['canceled']
        num_total = num_succeeded + num_processing + num_errored + num_expired + num_canceled
        assert num_total == len(batch_messages), f"Error: total number of requests {num_total} does not match the number of requests {len(batch_messages)}"
        tqdm_bar.n = num_succeeded + num_errored + num_expired + num_canceled
        tqdm_bar.total = num_total
        # tqdm_bar.set_postfix(completed=num_completed, total=num_total, failed=num_failed)
        tqdm_bar.set_postfix(
            completed=num_succeeded,
            processing=num_processing,
            errored=num_errored,
            expired=num_expired,
            canceled=num_canceled
        )
        if batch_status["status"] in ["completed", "errored", "expired", "canceled"]:
            tqdm_bar.close()
            break
        elif batch_status["status"] == "in_progress":
            tqdm_bar.desc = "In Progress"
            tqdm_bar.refresh()
        else:
            tqdm_bar.desc = batch_status["status"]
            tqdm_bar.refresh()
        time.sleep(random.randint(5, 10))
    
    results = get_batch_result(batch_result_id)
    return results


if __name__ == "__main__":
    from icecream import ic
    ic(call_worker_claude(["Hello", "Hi, I am claude", "What did I ask in the last response?"], "claude-3-opus-20240229"))


import os
import time
from typing import List
import google.ai.generativelanguage as glm
import google.generativeai as genai
from google.api_core.exceptions import ServiceUnavailable, ResourceExhausted
from .utils import with_timeout, decode_base64_image_url
genai.configure(api_key=os.environ.get("GEMINI_API_KEY"))
safety_settings = [
    {
        "category": "HARM_CATEGORY_DANGEROUS",
        "threshold": "BLOCK_NONE",
    },
    {
        "category": "HARM_CATEGORY_HARASSMENT",
        "threshold": "BLOCK_NONE",
    },
    {
        "category": "HARM_CATEGORY_HATE_SPEECH",
        "threshold": "BLOCK_NONE",
    },
    {
        "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
        "threshold": "BLOCK_NONE",
    },
    {
        "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
        "threshold": "BLOCK_NONE",
    },
]
# no image, multi-turn, do not use openai_generate, but can refer to it
def call_worker_gemini(messages:List[str], model_name, timeout:int=60, conv_system_msg=None, **generate_kwargs) -> str:
    # change messages to gemini format
    model = genai.GenerativeModel(model_name, system_instruction=conv_system_msg)
    
    new_messages = []
    role_map = {"user": "user", "assistant": "model"}
    for i, message in enumerate(messages):
        role = role_map[message["role"]]
        if isinstance(message["content"], str):
            new_messages.append({"role": role, "parts": [glm.Part(text=message["content"])]})
        elif isinstance(message["content"], list):
            parts = []
            for sub_message in message["content"]:
                if sub_message["type"] == "text":
                    parts.append(glm.Part(text=sub_message["text"]))
                elif sub_message["type"] == "image_url":
                    try:
                        image = decode_base64_image_url(sub_message["image_url"]['url'])
                    except Exception as e:
                        image = sub_message["image_url"]['url']
                    parts.append(image)
                else:
                    raise ValueError("Invalid message format")
            new_messages.append({"role": role, "parts": parts})
        else:
            raise ValueError("Invalid message format")
    
    stream = generate_kwargs.pop("stream", False)
    generation_config = genai.GenerationConfig(
        candidate_count=generate_kwargs.get("num_return_sequences", None),
        stop_sequences=generate_kwargs.get("stop", None),
        max_output_tokens=generate_kwargs.get("max_tokens", None),
        temperature=generate_kwargs.get("temperature", None),
        top_p=generate_kwargs.get("top_p", None),
        top_k=generate_kwargs.get("top_k", None),
        response_mime_type=generate_kwargs.get("response_mime_type", None),
        response_schema=generate_kwargs.get("response_schema", None),
    )
    request_options = genai.types.RequestOptions(
        timeout=timeout,
    )
    if "logprobs" in generate_kwargs:
        raise ValueError("logprobs is not supported in gemini")
    @with_timeout(timeout)
    def generate_content():
        return model.generate_content(new_messages, safety_settings=safety_settings, generation_config=generation_config, request_options=request_options, stream=stream)
    while True:
        try:
            response = generate_content()
            break
        except ServiceUnavailable as e:
            # sleep for a while and retry
            # print("ServiceUnavailable, retrying...")
            time.sleep(2)
            continue
        except ResourceExhausted as e:
            # sleep for a while and retry
            # print("ResourceExhausted, retrying...")
            time.sleep(10)
            continue
    try:
        if not stream:
            return response.text
        else:
            def generate_stream():
                for chunk in response:
                    yield chunk.text
            return generate_stream()
    except ValueError as e:
        print(f"Empty response from gemini due to {e}")
        return None

if __name__ == "__main__":
    from icecream import ic
    ic(call_worker_gemini(["Hello", "Hi, I am gemini", "What did I ask in the last response?"], "gemini-1.5-flash"))



import time
from llm_engines import LLMEngine
model_name="meta-llama/Meta-Llama-3-8B-Instruct"
llm = LLMEngine()
llm.load_model(
    model_name=model_name,
    num_workers=1, # number of workers
    num_gpu_per_worker=1, # tensor parallelism size for each worker
    engine="vllm", # or "sglang"
    use_cache=False
)
response = llm.call_model(model_name, "What is the capital of France?", temperature=0.0, max_tokens=None)
print(response)
llm.sleep_model(model_name)
time.sleep(20)
# wake up the model when calling it again if the model is in sleep mode
response = llm.call_model(model_name, "What is the capital of France?", temperature=0.0, max_tokens=None)

